package de.tu_darmstadt.tk.SmartHomeNetworkSim.view;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.event.MouseInputListener;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Link;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Packet;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;

/**
 * Listener which detects User Interaction with the {@link VisualisationPanel},
 * stores interaction information, triggers the controller and helps
 * visualisation of the interaction
 * 
 * @author Andreas T. Meyer-Berg
 */
public class VisualisationInteractor implements MouseInputListener, MouseMotionListener, ActionListener {

	/**
	 * Smart Home model which should be interacted with
	 */
	private Model model;

	/**
	 * Controller to notify when the model changes
	 */
	private Controller control;

	/**
	 * Panel which is observed
	 */
	private VisualisationPanel panel;
	
	/**
	 * Menu which is shown on right clicks
	 */
	private JPopupMenu rightClickMenu;
	
	/**
	 * RightClick MenuItem for SmartDevice creation**
	 * Finishes the drag operation, if a SmartDevice was dragged, repaint the panel and set dragged 
	 */
	private JMenuItem itemCreate;
	
	/**
	 * RightClick MenuItem for SmartDevice deletion
	 */
	private JMenuItem itemDelete;
	
	/**
	 * SmartDevice that is dragged on Screen
	 */
	public SmartDevice dragged = null;

	/**
	 * Current x Position of the dragged device
	 */
	public int dragged_x;

	/**
	 * Current y Position of the dragged device
	 */
	public int dragged_y;
	
	/**
	 * SmartDevice that was clicked
	 */
	private SmartDevice clicked = null;

	/**
	 * Creates a new VisualisationInteractor**
	 * Finishes the drag operation, if a SmartDevice was dragged, repaint the panel and set dragged 
	 *
	 * @param model
	 * @param controller
	 */
	public VisualisationInteractor(Model model, Controller controller, VisualisationPanel panel) {
		//Initialize the values
		this.model = model;
		this.control = controller;
		this.panel = panel;
		
		this.rightClickMenu = new JPopupMenu();
		itemCreate = new JMenuItem("Create SmartDevice");
		itemDelete = new JMenuItem("Delete");
		
		itemDelete.addActionListener(e -> {
			//Delete the clicked object
			controller.deleteSmartDevice(clicked);
			clicked = null;
			panel.repaint();
			});
		
		itemCreate.addActionListener(e->{
			//Print Links, Devices and Connections
			for(Link l:model.getConnectionNetworks())
				System.out.println("Link: "+l.getName());
			for(SmartDevice d: model.getDevices()){
				System.out.println("Device: "+d.getName());
				for(Connection c: d.getConnections()){
					if(c.getSource()==null)
						System.out.println("Connection: terminated");
					else
						System.out.println("Connection: "+c.getSource().getName());
				}
			}
			//Print Terminating Packages
			for(Connection c: model.getTerminatingConnections()){
				for(Packet p:c.getTerminationPackages(1000))
					System.out.println(p.toString());
			}
		});
		
		rightClickMenu.add(itemCreate);
		rightClickMenu.add(itemDelete);
		
		this.panel.add(rightClickMenu);
	}

	@Override
	public void mouseClicked(MouseEvent e) {
		if(dragged!=null){
			finishDrag();
		}
		/*
		 * FindSmartDevice that was clicked
		 */
		clicked = null;
		// Find the clicked SmartDevice
		for (SmartDevice d : model.getDevices()) {
			if (Math.abs(d.getX() - e.getX()) < panel.getVisualisationRadius()
					&& Math.abs(d.getY() - e.getY()) < panel.getVisualisationRadius()) {
				clicked = d;
			}
		}
		//If Rightclick
		if(e.getButton()==MouseEvent.BUTTON3) {
			//Show the RightClickMenu
			showRightClickMenu(clicked);	
		}
			
	}

	/**
	 * Shows the RightClick Menu on the Visualisation Panel, with Options for the given SmartDevice clickedOn, 
	 * if clickedOn is null, the RightClick Menu contains Options for creation of new SmartDevices.
	 * The Menu will be shown at the Mouse position on the VisualisationPanel. 
	 * @param clickedOn Device which was clicked, null if no
	 */
	protected void showRightClickMenu(SmartDevice clickedOn) {
		/**
		 * Mouse Position on the VisualisationPanel
		 */
		Point mousePos = panel.getMousePosition();
		//Just execute if Mouse Position is on the Panel
		if(mousePos != null){
			if(clickedOn == null){		
				itemCreate.setEnabled(true);
				itemDelete.setEnabled(false);
			}
			else{
				itemCreate.setEnabled(false);
				itemDelete.setEnabled(true);
			}
			//Show the RightClickMenu
			rightClickMenu.show(panel, mousePos.x, mousePos.y);
			rightClickMenu.setEnabled(true);
			rightClickMenu.setVisible(true);
		}
	}

	@Override
	public void mousePressed(MouseEvent e) {
		// Check if SmartDevice was clicked
		for (SmartDevice d : model.getDevices()) {
			if (Math.abs(d.getX() - e.getX()) < panel.getVisualisationRadius()
					&& Math.abs(d.getY() - e.getY()) < panel.getVisualisationRadius()) {
				// Recognize possible drag operation
				if (e.getButton() == MouseEvent.BUTTON1) {
					dragged = d;
					dragged_x = d.getX();
					dragged_y = d.getY();
				}
			}
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		// Finish drag operation
		if (e.getButton() == MouseEvent.BUTTON1)
			finishDrag();
	}

	@Override
	public void mouseEntered(MouseEvent e) {

	}

	@Override
	public void mouseExited(MouseEvent e) {

	}

	@Override
	public void actionPerformed(ActionEvent e) {

	}

	@Override
	public void mouseDragged(MouseEvent e) {
		// move dragged object on screen along y Axis
		if (e.getX() <= panel.getVisualisationRadius())
			dragged_x = panel.getVisualisationRadius();
		else if (e.getX() >= model.getWidth() - panel.getVisualisationRadius())
			dragged_x = model.getWidth() - panel.getVisualisationRadius();
		else
			dragged_x = e.getX();
		// move dragged object on screen along y Axis
		if (e.getY() <= panel.getVisualisationRadius())
			dragged_y = panel.getVisualisationRadius();
		else if (e.getY() >= model.getHeight() - panel.getVisualisationRadius())
			dragged_y = model.getHeight() - panel.getVisualisationRadius();
		else
			dragged_y = e.getY();
		// repaint the dragged smartDevice
		if (dragged != null)
			panel.repaint();
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		
	}
	
	/**
	 * Finishes the drag operation, if a SmartDevice was dragged, repaint the panel and set dragged to null
	 */
	public void finishDrag(){
		if (dragged != null && (dragged.getX() != dragged_x || dragged.getY() != dragged_y)) {
			control.moveSmartDevice(dragged, dragged_x, dragged_y, dragged.getZ());
			panel.repaint();
			dragged = null;
		}
	}
}
